<?php

if (!defined('ABSPATH') || !defined('AUTH_KEY')) {
    exit;
}

class TransacionCloudPaymentGateway extends WC_Payment_Gateway
{
    /**
     * @var TransactionCloudPayload
     */
    private $payloadService;

    /**
     * @var TransactionCloudApi
     */
    private $http;

    public function __construct()
    {
        $this->id = 'transaction_cloud';
        $this->icon = esc_url(plugin_dir_url( __DIR__ ) . '/assets/payment-methods.png');
        $this->has_fields = false;
        $this->method_title = 'TransactionCloud';
        $this->method_description = 'TransactionCloud integration for WooCommerce.';

        $this->init_form_fields();
        $this->init_settings();

        $this->title = $this->get_option('title');
        $this->description = $this->get_option('description');

        $this->initHttpClient();
        $this->initPayloadService();

        add_action('wp_enqueue_scripts', array($this, 'enqueueScripts'));

        add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));

        //https://your-domain.com/?wc-api=tc_webhook
        add_action('woocommerce_api_tc_webhook', array($this, 'processWebhook'));

        //https://your-domain.com/?wc-api=tc_return_url
        add_action('woocommerce_api_tc_return_url', array($this, 'handleReturnUrl'));

        add_action('wp_head', array($this, 'initVariables'));
    }

    public function initVariables()
    {
        $popupMode = $this->get_option('popup_mode') == 'yes' ? 'true' : 'false';

        echo '<script>var tc_params = {"popup_mode": '. $popupMode .'};</script>';
    }

    private function initPayloadService()
    {
        $this->payloadService = new TransactionCloudPayload();
        $this->payloadService->setCiphering('AES-128-CTR');
        $this->payloadService->setKey(AUTH_KEY);
    }

    private function initHttpClient()
    {
        $auth = new TransactionCloudAuth(
            $this->get_option('api_login'),
            $this->get_option('api_password'),
        );

        $this->http = new TransactionCloudApi();

        $this->http->addHeader(
            'TC-Version: v' . TC_WOO_VERSION
        );

        $this->http->useAuth($auth);

        $env = $this->get_option('environment');

        if ($env == 'production') {
            $this->http->useProductionEnvironment();
        }

        if ($env == 'beta') {
            $this->http->useBetaEnvironment();
        }
    }

    public function init_form_fields()
    {
        $this->form_fields = array(
            'enabled' => array(
                'title' => 'Enable/Disable',
                'type' => 'checkbox',
                'label' => 'Enable TransactionCloud',
                'default' => 'yes'
            ),
            'auto_configure' => array(
                'title' => 'Enable/Disable',
                'type' => 'checkbox',
                'label' => 'Auto configuration',
                'description' => 'Determines if security tag and product ID should be auto configured. In most cases should help, but in rare cases it is better to configure it manually.',
                'default' => 'yes'
            ),
            'title' => array(
                'title' => 'Title',
                'type' => 'text',
                'description' => 'This controls the title which the user sees during checkout',
                'default' => 'TransactionCloud',
                'desc_tip' => true,
            ),
            'description' => array(
                'title' => 'Description',
                'type' => 'textarea',
                'description' => 'This controls the description which the user sees during checkout',
                'default' => '',
                'desc_tip' => true,
            ),
            'api_login' => array(
                'title' => 'API Login',
                'type' => 'text',
                'description' => 'Enter your Transacion Cloud API Login',
                'default' => '',
                'desc_tip' => true,
            ),
            'api_password' => array(
                'title' => 'API Password',
                'type' => 'text',
                'description' => 'Enter your Transacion Cloud API password',
                'default' => '',
                'desc_tip' => true
            ),
            'security_tag' => array(
                'title' => 'Security Tag',
                'type' => 'text',
                'description' => 'TransactionCloud security tag sent as part of webhook request required to verify sender',
                'default' => '',
                'desc_tip' => true,
                // 'custom_attributes' => array(
                //     'readonly' => 'readonly'
                // )
            ),
            'product_id' => array(
                'title' => 'Product ID',
                'type' => 'text',
                'description' => 'Product ID used to initialize transacion',
                'default' => '',
                'desc_tip' => true,
                // 'custom_attributes' => array(
                //     'readonly' => 'readonly'
                // )
            ),
            'environment' => array(
                'title' => 'Environment',
                'type' => 'select',
                'description' => 'Environment',
                'default' => '',
                'desc_tip' => true,
                'options' => [
                    // 'beta' => 'Beta', // only for plugin develop purposes
                    'production' => 'Production',
                    'sandbox' => 'Sandbox',
                ]
            ),
            'popup_mode' => array(
                'title' => 'Pop-up mode',
                'type' => 'checkbox',
                'label' => 'Enable pop-up mode',
                'description' => 'Might be not available for all payments and modes',
                'default' => 'no'
            ),
            'clear_cart' => array(
                'title' => 'Cart',
                'type' => 'checkbox',
                'label' => 'Clear cart',
                'description' => 'Whether cart should be cleared after checkout. In most cases this option should be checked.',
                'default' => 'yes'
            )
        );
    }

    public function process_admin_options()
    {
        $return = parent::process_admin_options(); // this order is important -> we first need to save data from user and then try to get additional data from API

        if ($this->get_option('auto_configure') === 'yes') {
            $configure = new TransactionCloudConfigure();
            $configure->setUrl(esc_url(home_url()));

            $auth = new TransactionCloudAuth(
                $this->get_option('api_login'),
                $this->get_option('api_password'),
            );

            $env = $this->get_option('environment');

            if ($env == 'production') {
                $this->http->useProductionEnvironment();
            }

            if ($env == 'beta') {
                $this->http->useBetaEnvironment();
            }

            if ($env == 'sandbox') {
                $this->http->useSandboxEnvironment();
            }

            $this->http->useAuth($auth);

            try {
                $config = $this->http->configure($configure);

                $this->update_option('security_tag', $config->securityTag);
                $this->update_option('product_id', $config->productId);
            } catch (\Exception $e) {
                $this->update_option('security_tag', '');
                $this->update_option('product_id', '');
                $this->add_error($e->getMessage());
            }
        }

        return $return;
    }

    public function handleReturnUrl()
    {
        if (!isset($_GET['id'])) {
            status_header(400);
            exit('Missing transaction ID');
        }

        $id = $_GET['id'];

        try {
            $transaction = $this->http->getTransaction($id);
            $payload = $transaction->payload;
        } catch (\Exception $e) {
            status_header(500);
            exit($e->getMessage());
        }

        if (!$payload) {
            status_header(404);
            exit('Not found');
        }

        try {
            $orderId = $this->payloadService->decode($payload);
        } catch (\Exception $e) {
            status_header(500);
            exit($e->getMessage());
        }

        $order = wc_get_order($orderId);

        if (!$order) {
            status_header(404);
            exit('Not found');
        }

        wp_redirect($order->get_checkout_order_received_url());
        exit();
    }

    public function enqueueScripts()
    {
        $env = $this->get_option('environment');
        $pluginPath = plugin_dir_url(__DIR__);

        if ($env == 'beta') {
            wp_enqueue_script('tc-widget', 'https://beta-cdn.transaction.cloud/1.0.2/widget.development.js', array('jquery'), TC_WOO_VERSION, true);
        }

        if ($env == 'production') {
            wp_enqueue_script('tc-widget', 'https://cdn.transaction.cloud/1.0.2/widget.min.js', array('jquery'), TC_WOO_VERSION, true);
        }

        if ($env == 'sandbox') {
            wp_enqueue_script('tc-widget', 'https://sandbox-cdn.transaction.cloud/1.0.2/widget.sandbox.min.js', array('jquery'), TC_WOO_VERSION, true);
        }

        wp_enqueue_script('tc-main', $pluginPath . 'assets/js/transaction-cloud.js', array('jquery', 'tc-widget'), TC_WOO_VERSION, true);
    }

    public function process_payment($order_id)
    {
        $order = wc_get_order($order_id);

        try {
            $payloadData = json_encode(
                array(
                    'order_id' => $order_id,
                    'home_url' => home_url()
                ),
                JSON_THROW_ON_ERROR
            );

            $payload = $this->payloadService->encode($payloadData);
        } catch (TransactionCloudPayloadEncodeException $e) {
            wc_add_notice($e->getMessage(), 'error');
            return false;
        }

        // Prepare the data to send to the TransactionCloud endpoint.
        $data = array(
            'amount' => $order->get_total(),
            'currency' => $order->get_currency(),
            'payload' => $payload,
            'email' => $order->get_billing_email(),
            'firstname' => $order->get_billing_first_name(),
            'lastname' => $order->get_billing_last_name()
        );

        // Send a request to the TransactionCloud endpoint to get the payment URL.
        try {
            $customizedProduct = $this->getCustomizedProduct($data);
            $paymentUrl = $customizedProduct->link;
        } catch (\Exception $e) {
            $error = sprintf('%s (%s)', $e->getMessage(), $this->get_option('product_id'));
            wc_add_notice($error, 'error');
            return false;
        }

        // Reduce stock levels.
        wc_reduce_stock_levels($order_id);

        // Empty cart
        if ($this->get_option('clear_cart') === 'yes') {
            wc_empty_cart();
        }

        // Set the order status to pending.
        $order->update_status('pending', 'Awaiting TransactionCloud payment.');

        // Add a note to the order.
        $order->add_order_note('Redirecting to TransactionCloud for payment.');

        // Redirect to the payment URL in a new tab.
        return array(
            'result' => 'success',
            'redirect' => $paymentUrl,
            'customProductId' => $customizedProduct->customProductId
        );
    }

    private function getCustomizedProduct($data)
    {
        $productId = $this->get_option('product_id');
        $customizeProduct = new TransactionCloudCustomizeProduct($productId);

        $price = new TransactionCloudPrice();
        $price->setCurrency($data['currency']);
        $price->setValue(floatval($data['amount']));

        $customizeProduct->addPrice($price);
        $customizeProduct->setPayload($data['payload']);
        $customizeProduct->setUserEmail($data['email']);
        $customizeProduct->setUserFirstName($data['firstname']);
        $customizeProduct->setUserLastName($data['lastname']);

        $customizedProduct = $this->http->customizeProduct($customizeProduct);

        if (!$customizedProduct) {
            return null;
        }

        return $customizedProduct;
    }

    public function processWebhook()
    {
        $data = json_decode(file_get_contents('php://input'), true);

        if (!$this->isSecurityTagValid()) {
            status_header(400);
            exit('Invalid webhook signature.');
        }

        if (!$this->shouldProcessWebhook($data)) {
            status_header(202);
            exit('Webhook skipped');
        }

        $status = $data['transactionStatus'];

        try {
            $payloadData = json_decode($this->payloadService->decode($data['payload']), true);
        } catch (TransactionCloudPayloadDecodeException $e) {
            status_header(400);
            exit($e->getMessage());
        }

        if (!isset($payloadData['order_id']) || !isset($payloadData['home_url'])) {
            status_header(400);
            exit('Invalid webhook payload.');
        }

        $homeUrl = $payloadData['home_url'];

        if ($homeUrl && $homeUrl != home_url()) {
            status_header(202);
            exit('Home url missmatch');
        }

        $orderId = $payloadData['order_id'];

        if (!$orderId) {
            status_header(400);
            exit('Invalid webhook payload.');
        }

        $order = wc_get_order($orderId);

        if (!$order) {
            status_header(404);
            exit('Order not found.');
        }

        if ($status == 'ONE_TIME_PAYMENT_STATUS_PAID') {
            $order->update_status('completed', 'TransactionCloud payment completed.');
        }

        if ($status == 'TRANSACTION_STATUS_REFUNDED') {
            $order->update_status('refunded', 'TransactionCloud refunded.');
        }

        status_header(200);

        exit('Webhook processed successfully.');
    }

    private function shouldProcessWebhook($data)
    {
        $requiredFields = [
            'payload',
            'transactionStatus',
            'webhookType'
        ];

        foreach ($requiredFields as $field) {
            if (!isset($data[$field])) {
                return false;
            }
        }

        $webhookType = $data['webhookType'];

        if ($webhookType != 'ONETIME_PAID' && $webhookType != 'ONETIME_REFUNDED') {
            return false;
        }

        return true;
    }

    /**
     * Checks whether webhook is sent by TransactionCloud. If security tag
     * is not provideed in config returns true.
     *
     * @return bool
     */
    private function isSecurityTagValid()
    {
        $securityTag = $this->get_option('security_tag');

        if (!$securityTag) {
            return true;
        }

        $requestSecurityTag = TransacionCloudUtil::getRequestHeader('X-Transaction-Cloud-Security-Tag');

        if (!$requestSecurityTag) {
            return false;
        }

        return $securityTag == $requestSecurityTag;
    }
}
